import unittest
from unittest.mock import patch

from tmc import points
from tmc.utils import load, load_module, reload_module, get_stdout, check_source
from functools import reduce
import os
import os.path
import textwrap
from random import choice, randint
from string import ascii_lowercase, digits

exercise = 'src.password_generator_part_2'
function = "generate_strong_password"
punctuation = "!?=+-()#"

def chars_ok(s: str, g: str):
    return len([x for x in s if x not in g]) == 0

def contains(s: str, g: str):
    return any([x in s for x in g])

@points('1.password_generator_part_2')
class PasswordGenerator2Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with patch('builtins.input', side_effect=[AssertionError("Att be om input från användaren förväntades inte")]):
           cls.module = load_module(exercise, 'en')

    def test_0a_main_program_ok(self):
        ok, line = check_source(self.module)
        message = """Koden för att testa funktionerna ska placeras inuti
if __name__ == "__main__":
block. Följande rad ska flyttas:
"""
        self.assertTrue(ok, message+line)

    def function_exists(self):
        try:
            from src.password_generator_part_2 import generate_strong_password
        except:
            self.assertTrue(False, "Din kod bör innehålla funktionen generate_strong_password(length: int, numbers: bool, special_characters: bool)")

    def test2_type_of_return_value(self):
        try:
            from src.password_generator_part_2 import generate_strong_password
            val = generate_strong_password(2,False,False)
            taip = str(type(val)).replace("<class '","").replace("'>","")
            self.assertTrue(type(val) == str, 
                f"Funktionen generate_strong_password ska returnera en string. Istället för {val} vilket är en {taip} när parameter (2, False, False) används.")
        except:
            self.assertTrue(False, f"Det uppstod ett fel när funktionen anropades med parametervärdena (2, False, False))")

    def test3_uses_import_expressions(self):
        with open("src/password_generator_part_2.py") as f:
            cont = f.read()
            self.assertTrue("import" in cont and "random" in cont, 
                f"Ditt program importerar inte random bibliotek med importuttrycket.")

    def test4_test_only_letters(self):
        test_cases = [5,6,8,12,14,16]
        for test_case in test_cases:
            with patch('builtins.input', side_effect=[AssertionError("Att be om input från användaren förväntades inte")]):
                reload_module(self.module)
                generate_strong_password = load(exercise, function, 'en')

                val1 = generate_strong_password(test_case, False, False)
                val2 = generate_strong_password(test_case, False, False)

                self.assertTrue(len(val1) == test_case, f"Lösenordets längd är {len(val1)}, men det bör vara {test_case} när parametern är {test_case}: {val1}")
                self.assertTrue(chars_ok(val1, ascii_lowercase), f"Lösenordet innehåller andra tecken än tillåtna: \n{val1} \nnär parametrarna är det {test_case, False, False}")
                self.assertTrue(chars_ok(val2, ascii_lowercase), f"Lösenordet innehåller andra tecken än tillåtna: \n{val2} \nnär parametrarna är det {test_case, False, False}") 
                self.assertNotEqual(val1, val2, f"Att anropa funktionen returnerar samma lösenord varje gång: \n{val1} \nnär parametern är {test_case}")

    def test5_test_letters_and_numbers(self):
        test_cases = [5,6,8,12,14,16]
        for test_case in test_cases:
            with patch('builtins.input', side_effect=[AssertionError("Att be om input från användaren förväntades inte")]):
                reload_module(self.module)
                generate_strong_password = load(exercise, function, 'en')

                val1 = generate_strong_password(test_case, True, False)
                val2 = generate_strong_password(test_case, True, False)

                self.assertTrue(len(val1) == test_case, f"Lösenordets längd är {len(val1)}, men det bör vara {test_case} när parametern är {test_case}: {val1}")
                self.assertTrue(chars_ok(val1, ascii_lowercase + digits), f"Lösenordet innehåller andra tecken än tillåtna: \n{val1} \nnär parametrarna är det {test_case, True, False}")
                self.assertTrue(chars_ok(val2, ascii_lowercase + digits), f"Lösenordet innehåller andra tecken än tillåtna: \n{val2} \nnär parametrarna är det {test_case, True, False}") 
                self.assertTrue(contains(val1, digits), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{digits}': {val1} när parametrarna är {test_case, True, False} ")
                self.assertTrue(contains(val1, ascii_lowercase), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{ascii_lowercase}': {val1} när parametrarna är {test_case, True, False} ") 
                self.assertNotEqual(val1, val2, f"Att anropa funktionen returnerar samma lösenord varje gång: {val1} när parametrarna är {test_case, False, True}")

    def test6_test_letters_and_special_characters(self):
        test_cases = [5,6,8,12,14,16]
        for test_case in test_cases:
            with patch('builtins.input', side_effect=[AssertionError("Att be om input från användaren förväntades inte")]):
                reload_module(self.module)
                generate_strong_password = load(exercise, function, 'en')

                val1 = generate_strong_password(test_case, False, True)
                val2 = generate_strong_password(test_case, False, True)

                self.assertTrue(len(val1) == test_case, f"Lösenordets längd är {len(val1)}, men det bör vara {test_case} när parametern är {test_case}: {val1}")
                self.assertTrue(chars_ok(val1, ascii_lowercase + punctuation), f"Lösenordet innehåller andra tecken än tillåtna: \n{val1} \nnär parametrarna är det {test_case, False, True}")
                self.assertTrue(chars_ok(val2, ascii_lowercase + punctuation), f"Lösenordet innehåller andra tecken än tillåtna: \n{val2} \nnär parametrarna är det {test_case, False, True}") 
                self.assertTrue(contains(val1, ascii_lowercase), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{ascii_lowercase}': {val1} när parametrarna är {test_case, False, True} ") 
                self.assertTrue(contains(val1, punctuation), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{punctuation}': {val1} när parametrarna är {test_case, False, True} ")
                self.assertNotEqual(val1, val2, f"Att anropa funktionen returnerar samma lösenord varje gång: {val1} när parametrarna är {test_case, False, True}")

    def test7_test_all(self):
        test_cases = [5,6,8,12,14,16]
        for test_case in test_cases:
            with patch('builtins.input', side_effect=[AssertionError("Att be om input från användaren förväntades inte")]):
                reload_module(self.module)
                generate_strong_password = load(exercise, function, 'en')

                val1 = generate_strong_password(test_case, True, True)
                val2 = generate_strong_password(test_case, True, True)

                self.assertTrue(len(val1) == test_case, f"Lösenordets längd är {len(val1)}, men det bör vara {test_case} när parametern är {test_case}: {val1}")
                self.assertTrue(chars_ok(val1, ascii_lowercase + punctuation + digits), f"Lösenordet innehåller andra tecken än tillåtna: \n{val1} \nnär parametrarna är det {test_case, True, True}")
                self.assertTrue(chars_ok(val2, ascii_lowercase + punctuation + digits), f"Lösenordet innehåller andra tecken än tillåtna: \n{val2} \nnär parametrarna är det {test_case, True, True}") 
                self.assertTrue(contains(val1, punctuation), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{punctuation}': {val1} när parametrarna är {test_case, True, True} ")
                self.assertTrue(contains(val1, digits), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{digits}': {val1} när parametrarna är {test_case, True, True} ") 
                self.assertTrue(contains(val1, ascii_lowercase), f"Lösenordet innehåller inte ett enda tecken från uppsättningen '{ascii_lowercase}': {val1} när parametrarna är {test_case, True, True} ") 
                self.assertNotEqual(val1, val2, f"Att anropa funktionen returnerar samma lösenord varje gång: {val1} när parametrarna är {test_case, True, True}")

if __name__ == '__main__':
    unittest.main()